#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <math.h>
#include <hdf5.h

/* :: -------------------------------------------- ::  */
/*    DATA SEGMENT                                     */

                                                                         /* :: -------------------------------------------- ::  */
                                                                         /*     variables related to reading the directory      */
int WalCool_get_redshift_table(void);
int is_it_a_tablefile(char *, int);
int compare_redshifts(const void *, const void *);

struct my_direntry
{
  char   redshift_str[10];
  double redshift;
};

static DIR    *WalCool_CoolTables_dir;
static struct my_direntry *WalCool_CoolTables_redshifts;
static int    WalCool_CoolTables_num;


                                                                         /* :: -------------------------------------------- ::  */
                                                                         /*     define tables                                   */
static int SupCoolTable, InfCoolTable;
static int WalCool_n[6];
#define WalCool_n_Rho WalCool_n[0]                                       /*!< number of density bins      */
#define WalCool_n_T   WalCool_n[1]                                       /*!< number of temperature bins  */
#define WalCool_n_Hef WalCool_n[2]                                       /*!< number of helium fractions  */
#define WalCool_n_El  WalCool_n[3]                                       /*!< number of elements          */
#define WalCool_n_Ab  WalCool_n[4]                                       /*!< number of solar abundances  */

static double *WalCool_arrays;
static double *WalCool_Rho;                                              /*!< stores the density bins        */
static double *WalCool_T;                                                /*!< stores the temperature bins    */
static double *WalCool_U;                                                /*!< stores the energy density bins */
static double *WalCool_Hef;                                              /*!< stores the Helium fractions    */
static double *WalCool_Ab;                                               /*!< stores the solar abundances    */
static double *WalCool_enh;                                              /*!< stores the electron abundances over n_H and */
static double *WalCool_mu;                                               /*   mean molecular mass as a function of n_H, T */
                                                                         /*   and n_He/n_H*/


#define ENH_MU_IDX(T, nH, Hef) ((nH) * (WalCool_n_T * WalCool_n_Hef) + (Hef) * WalCool_n_T + (T))

static char **WalCool_names_tab;
static char *WalCool_names, *WalCool_El_names, *WalCool_Ab_names;


static float *WALCOOLTABLES, *WalMfreeCoolTables, *WalCooltables;

#define METFREE_COOLRATE_IDX(T_idx, R_idx, He_idx, Redsh_idx) ((T_idx) * ((WalCool_n_Hef * 2) * WalCool_n_Rho) + (R_idx) * ((WalCool_n_Hef * 2)) + (He_idx) * 2 + (Redsh_idx))
#define COOLRATE_IDX(T_idx, R_idx, el_idx, Redsh_idx) ((T_idx) * ((WalCool_n_El * 2) * WalCool_n_Rho) + (R_idx) * ((WalCool_n_El * 2)) + (el_idx) * 2 + (Redsh_idx))


                                                                         /* :: -------------------------------------------------- ::  */
                                                                         /*     variables and routines related to reading tables      */

int get_WalCoolTable(double);





/* :: -------------------------------------------- ::  */
/*    CODE SEGMENT                                     */



int WalCool_get_redshift_table(void)
{

  int n;
  struct dirent *dentry;


                                                                         /* as first, task 0 gets the content of the directory */
                                                                         /* we DO NOT use scandir because it uses the system's */
                                                                         /* malloc routine and we want to stick onto mymalloc  */
                                                                         /* (isnt'it ?)                                        */

  if( (WalCool_CoolTables_dir = opendir(WalCool_CoolTables_path)) == NULL)
    {
      printf("The directory %s that should contain the cooling tables from Wiersma et al. does not exist\n");
      exit(9191000);
    }
  
  n = 0;                                                                 /* this first cycle is needed to determine how many   */
                                                                         /* tables are in the directory                        */
  while(dentry = readdir(WalCool_CoolTables_dir))
    if( is_it_a_tablefile(dentry->d_name, -1) )
      n++;

  WalCool_CoolTables_num = n;
  
  closedir(WalCool_CoolTables_dir);
                                                                         /* re-open the directory */
  WalCool_CoolTables_dir = opendir(Wal_CoolTables_dir_name);

  WalCool_CoolTables_redshifts = malloc(sizeof(struct my_direntry) * n); /* allocate memory to store the entries  */

  n = 0;
  while(dentry = readdir(WalCool_CoolTables_dir))
    if( is_it_a_tablefile(dentry->d_name, n) )
      n++;
    
  closedir(WalCool_CoolTables_dir);

  qsort(WalCool_CoolTables_redshifts, WalCool_CoolTables_num, sizeof(struct my_direntry), compare_redshifts);
  
  return WalCool_CoolTables_num;
}



int is_it_a_tablefile(char *name, int n)
/* a file is a table file whether its name is made by the following pattern */
/*      z_[:digit:].[:digit:].hdf5                                          */
/* where [:digit:] means un unknown number od digits                        */
{
  int  pos, dotpos, ndots;
  
  pos = strlen(name);

  if(pos < 10)                                                           /* check for the minimum expected lenght */
    return 0;

  if(strcmp(&name[pos - 5], ".hdf5") != 0x0)                             /* check whether the suffix is ".hdf5" */
    return 0;

  if(strncmp(name, "z_", 2) != 0x0)                                      /* check whether the prefix is "z_" */
    return 0;

  dotpos = pos-5;                                                        /* points to the '.' */

  for(ndots = 0, pos = 2; pos < dotpos; pos++)
    {      
      if(!isdigit(name[pos]))
        {
          if(name[pos] == '.')
            {
              if(ndots == 1)
                return 0;
              else
                ndots = 1;
            }
          else
            return 0;
        }
    }

  if(n >= 0)
    {
      strncpy(WalCool_CoolTables_redshifts[n].redshift_str, &name[2], dotpos-2);
      WalCool_CoolTables_redshifts[n].redshift_str[dotpos-2]='\0';
      WalCool_CoolTables_redshifts[n].redshift = atof(WalCool_CoolTables_redshifts[n].redshift_str);
    }
  
  return 1;
}

int compare_redshifts(const void *A, const void *B)
{
  if( (*(struct my_direntry*)A).redshift > (*(struct my_direntry*)B).redshift)
    return 1;
  if( (*(struct my_direntry*)A).redshift < (*(struct my_direntry*)B).redshift)
    return -1;
  return 0;
}



void WalCool_get_table(int redshift_index, int tab_index)
{
  char   filename[300], dname;
  hid_t  fid, dset;
  herr_t hd_report;

  float  *rtable;
  int    element, Ridx, Tidx;
  
  sprintf(filename, "%s/z_%s.hdf5", WalCool_CoolTables_path, WalCool_CoolTables_redshifts[redshift_index].redshift_str);
  fid = H5Fopen(filename, H5F_ACC_RDONLY, H5P_DEFAULT);

                                                                         /* get cooling from metal-free: H + He */
  
  rtable = (float*)mymalloc("temporary_table", sizeof(float) * WalCool_n_T * WalCool_n_Hef);

  sprintf(dname, "/Metal_free/Net_Cooling");
  dset      = H5Dopen(fid, dname);
  hd_report = H5Dread(dset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, rtable);
  hd_report = H5Dclose(dset);
  
  for(Ridx = 0; Ridx < WalCool_n_Rho; Ridx++)
    for(Tidx = 0; Tidx < WalCool_n_T; Tidx++)
      SET_METFREE_COOLRATE(Tidx, Ridx, tab_index, rtable[Tidx * WalCool_n_T + Ridx]);

  myfree(rtable);
  rtable = (float*)mymalloc("temporary_table", sizeof(float) * WalCool_n_T * WalCool_n_T);
                                                                         /* get cooling from metals */
  for(element = 0; element < WalCool_n_El; element++)
    {
      sprintf(dname, "/%s/Net_Cooling", WalCool_El_names[element]);
      dset      = H5Dopen(fid, dname);
      hd_report = H5Dread(dset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, rtable);
      hd_report = H5Dclose(dset);
      
      for(Ridx = 0; Ridx < WalCool_n_Rho; Ridx++)
        for(Tidx = 0; Tidx < WalCool_n_T; Tidx++)
          SET_COOLRATE(Tidx, Ridx, element, tab_index, rtable[Tidx * WalCool_n_T + Ridx]);
    }

  return;  
}


void WalCool_get_collis_table()
{
  
}

void WalCool_Initialize_get_header(void)
{
  char   filename[300];
  hidt_t fid, dataset;
  herr_t hd_report;

  int    dimensions[6], k = 0;
  
  if(ThisTask == 0)
    {
      sprintf(filename, "%s/z_%s.hdf5", WalCool_CoolTables_path, WalCool_CoolTables_redshifts[WalCool_Cooltables_num].redshift_str);
      fid = H5Fopen(filename, H5F_ACC_RDONLY, H5P_DEFAULT);

                                                                         /* get the arrays dimensions */
      
      dataset = H5Dopen(fid, "/Header/Number_of_density_bins");
      hd_report = H5Dread(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &WalCool_n_Rho);
      hd_report = H5Dclose(dataset);
      dimensions[k++] = WalCool_n_Rho;

      dataset = H5Dopen(fid, "/Header/Number_of_temperature_bins");
      hd_report = H5Dread(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &WalCool_n_T);
      hd_report = H5Dclose(dataset);
      dimensions[k++] = WalCool_n_T;
      
      dataset = H5Dopen(fid, "/Header/Helium/Number_of_helium_fractions");
      hd_report = H5Dread(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &WalCool_n_Hef);
      hd_report = H5Dclose(dataset);
      dimensions[k++] = WalCool_n_Hef;

      dataset = H5Dopen(fid, "/Header/Number_of_species");
      hd_report = H5Dread(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &WalCool_n_El);
      hd_report = H5Dclose(dataset);
      dimensions[k++] = WalCool_n_El;

      dataset = H5Dopen(fid, "/Header/Abundances/Number_of_abundances");
      hd_report = H5Dread(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &WalCool_n_Ab);
      hd_report = H5Dclose(dataset);
      dimensions[k++] = WalCool_n_Ab;
    }

  MPI_Bcast(dimensions, 6, MPI_INT, 0, MPI_COMM_WORLD);

  if(ThisTask != 0)
    {
      k = 0;
      WalCool_n_Rho = dimensions[k++];
      WalCool_n_T   = dimensions[k++];
      WalCool_n_Hef = dimensions[k++];
      WalCool_n_El  = dimensions[k++];
      WalCool_n_Ab  = dimensions[k++];
    }

                                                                         /* allocate memory for array */

  WalCool_arrays = (float *) mymalloc("WalCool_arrays", (WalCool_n_Rho   +    /* density bins */
                                                         2 * WalCool_n_T +    /* T and U bins */
                                                         WalCool_n_Hef   +    /* Helium fractions */
                                                         2 * WalCool_n_Hef * WalCool_n_T * WalCool_n_Rho + /* mean molecular mass and electrons over n_H */
                                                         WalCool_n_Ab) *      /* solar abundances */
                                      sizeof(float));

  WalCool_Rho = WalCool_arrays;
  WalCool_T   = WalCool_Rho + WalCool_n_Rho;
  WalCool_U   = WalCool_T   + WalCool_n_T;
  WalCool_Hef = WalCool_U   + WalCool_n_T;
  WalCool_Ab  = WalCool_Hef + WalCool_n_Hef;
  WalCool_enH = WalCool_Ab  + WalCool_n_Ab;
  WalCool_mu  = WalCool_enH + WalCool_n_Rho * WalCool_n_T * WalCool_n_Hef;
  
                                                                         /* get the arrays */
  if(ThisTask == 0)
    {
                                                                         /* - solar abundances by mass */
      dataset     = H5Dopen(fid, "/Header/Abundances/Solar_mass_fractions");
      hd_report   = H5Dread(dataset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, WalCool_Ab);
      hd_report   = H5Dclose(dataset);


                                                                         /* - temperature bins */
      dataset     = H5Dopen(fid, "/Metal_free/Temperature_bins");
      hd_report   = H5Dread(dataset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, WalCool_T);
      hd_report   = H5Dclose(dataset);
      
                                                                         /* - density bins */
      dataset     = H5Dopen(fid, "/Metal_free/Hydrogen_density_bins");
      hd_report   = H5Dread(dataset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, WalCool_Rho);
      hd_report   = H5Dclose(dataset);


                                                                         /* - helium fraction bins */
      dataset     = H5Dopen(fid, "/Metel_free/Helium_mass_fraction_bins");
      hd_report   = H5Dread(dataset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, WalCool_Hef);
      hd_report   = H5Dclose(dataset);


                                                                         /* - mean particle mass */
      dataset     = H5Dopen(fid, "/Metel_free/Mean_particle_mass");
      hd_report   = H5Dread(dataset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, WalCool_mu);
      hd_report   = H5Dclose(dataset);


                                                                         /* - elecron density over n_h */
      dataset     = H5Dopen(fid, "/Metel_free/Eectron_density_over_n_h");
      hd_report   = H5Dread(dataset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, WalCool_enh);
      hd_report   = H5Dclose(dataset);

      
                                                                         /* - energy density bins */
      dataset     = H5Dopen(fid, "/Metal_free/Temperature/Energy_density_bins");
      hd_report   = H5Dread(dataset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, WalCool_U);
      hd_report   = H5Dclose(dataset);

    }

  MPI_Bcast(WalCool_arrays, (WalCool_n_Rho + WalCool_n_T*2 + WalCool_n_Hef + WalCool_n_Ab +
                             WalCool_n_Rho * WalCool_n_T * WalCool_n_Hef) * sizeof(float), MPI_BYTE, 0, MPI_COMM_WORLD);
  
                                                                         /* allocate memory for names */

  WalCool_names_tab = (char**)mymalloc("WalCool_names_tab", (WalCool_n_Hef + WalCool_n_El + WalCool_n_Ab) * sizeof(char*));
  WalCool_names     = (char*) mymalloc("WalCool_names",     (WalCool_n_Hef + WalCool_n_El + WalCool_n_Ab) * NAME_LEN * sizeof(char));

  WalCool_Hef_names = WalCool_names_tab;
  for (pos = i = 0; i < WalCool_n_Hef; i++, pos += NAME_LEN)
    WalCool_Hef_names[i] = WalCool_names[pos];
  
  WalCool_El_Names = WalCool_He_names + WalCool_n_Hef;
  for (i = 0; i < WalCool_n_El; i++, pos += NAME_LEN)
    WalCool_El_names[i] = WalCool_names[pos];

  WalCool_Ab_Names = WalCool_El_names + WalCool_n_El;
  for (i = 0; i < WalCool_n_Ab; i++, pos += NAME_LEN)
    WalCool_Ab_names[i] = WalCool_names[pos];

                                                                         /* get the names */
  if(ThisTask == 0)
    {
      datatype = H5Tcopy(H5T_C_S1);
      hd_report = H5Tset_size(datatype, H5T_VARIABLE);

                                                                         /* - He fraction names */
      dataset = H5Dopen(fid, "/Header/Helium_names");
      hd_report = H5Dread(dataset, datatype, H5S_ALL, H5S_ALL, H5P_DEFAULT, WalCool_Hef_names);  
      hd_report = H5Dclose(dataset);

                                                                         /* - elements names */
      dataset = H5Dopen(fid, "/Header/Species_names");
      hd_report = H5Dread(dataset, datatype, H5S_ALL, H5S_ALL, H5P_DEFAULT, WalCool_El_names);  
      hd_report = H5Dclose(dataset);

                                                                         /* - abundances names */
      dataset = H5Dopen(fid, "/Header/Abund_names");
      hd_report = H5Dread(dataset, datatype, H5S_ALL, H5S_ALL, H5P_DEFAULT, WalCool_Ab_names);  
      hd_report = H5Dclose(dataset);

      /* close the file */
      hd_report = H5Fclose(fid);
    }

  MPI_Bcast((WalCool_names, WalCool_n_Hef + WalCool_n_El + WalCool_n_Ab) * NAME_LEN, MPI_BYTE, 0, MPI_COMM_WORLD);
  
  MPI_Barrier(MPI_COMM_WORLD);
  
  return;
}


void WalCool_Initialize(void)
{
                                                                         /* get general informations from headers */
  WalCool_Initialize_get_header();

                                                                         /* allocate space for tables */
  WALCOOLTABLES = (float*)mymalloc("WalCoolTables", sizeof(float) * 2 * (WalCool_n_El + WalCool_n_Hef) * WalCool_n_Rho * WalCool_n_T);
  
                                                                         /* set pointers to H+He cooling */
  WalMfreeCoolTables = WALCOOLTABLES;
                                                                         /* set pointer to metal cooling tables */
  WalCooltables = WALCOOLTABLES + 2 * WalCool_n_Hef * WalCool_n_Rho * WalCool_n_T);
  
  return;
}

int main(int argc, char **argv)
{

  int i;
  
  WalCool_get_redshift_table(*(argv + 1));

  for(i = 0; i < WalCool_CoolTables_num; i++)
    printf("z_%s.hdf5 %7.4g\n", WalCool_CoolTables_redshifts[i].redshift_str, WalCool_CoolTables_redshifts[i].redshift);

  free(WalCool_CoolTables_redshifts);
  
  return 0;
}
